"""
This module brings together all evaluation functionalities.
"""

from ._folding import *
from ._sampling import *
from ._evaluation import *
from ._functions import *
from ._processing import *
from ._parallelization import *



"""
This module implements the SMOTE_AMSR method.
"""

import numpy as np
from sklearn.ensemble import IsolationForest

from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['SMOTE_AMSR']

class SMOTE_AMSR(OverSampling):
    """
    Description:
        combing the attention mechanism of sparse regions and the 
        interpolation technique of geometric topology to Generate 
        New Minority Data. 
    """
    categories = [OverSampling.cat_extensive]

    # @_deprecate_positional_args
    def __init__(self,
                 proportion=1.0,
                 *,
                 topology='mesh',
                 random_state=None):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            topology (str): the topology to use ('bus'/'star'/'mesh')
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """

        super().__init__(random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0.0)
        self.check_isin(topology, "topology", ['star', 'bus', 'mesh'])

        self.proportion = proportion
        self.topology = topology

    @classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.
        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'topology': ['bus', 'star', 'mesh'],}

        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def _star_generator(self, X, n_to_sample, scores, classifier):
        """
        Make samples with star topology

        Args:
            X (np.array): the feature vectory
            n_to_sample (int): the number of samples to generate
            scores (np.array): classification scores
            classifier (obj): the classifier

        Returns:
            np.array: the generated feature vectors
        """
        #samples = []

        X_mean = np.mean(X, axis=0)
        mean_anomaly = classifier.decision_function(X_mean.reshape(1,-1))[0]
        #k = max([1, int(np.rint(n_to_sample/len(X)))])
        max_min = max(max(scores), mean_anomaly) - min(min(scores), mean_anomaly)

        # this array stores all generated samples
        samples = np.zeros(shape=(n_to_sample, X.shape[1]))

        # the base points
        sample_indices = self.random_state.choice(X.shape[0], n_to_sample, replace=True)
        base_points = X[sample_indices]

        # differences
        diffs = X_mean - base_points

        # the anomaly greater or equal / less masks
        greater_equal_mask = mean_anomaly >= scores[sample_indices]
        less_mask = ~greater_equal_mask

        # the greater or equal case vectorized
        d_ge = mean_anomaly - scores[sample_indices][greater_equal_mask]

        if max_min > 0:
            d_ge = d_ge / max_min

        steps_ge = self.random_state.uniform(size=(np.sum(greater_equal_mask), X.shape[1]))
        steps_ge = (steps_ge.T * (1.0 - d_ge) + d_ge).T

        samples[greater_equal_mask] = base_points[greater_equal_mask] + steps_ge * diffs[greater_equal_mask]

        # the less case vectorized
        d_less = scores[sample_indices][less_mask] - mean_anomaly

        if max_min > 0:
            d_less = d_less / max_min

        steps_less = self.random_state.uniform(size=(np.sum(less_mask), X.shape[1]))
        steps_less = (steps_less.T * (1.0-d_less)).T

        samples[less_mask] = base_points[less_mask] + steps_less * diffs[less_mask]

        #for ix in range(len(X)):
        #    diff = X_mean - X[ix]
        #    if mean_amomaly>=scores[ix]:
        #        d = mean_amomaly-scores[ix] / max_min
        #        step = self.random_state.uniform(low=0+d, high=1.0, size=(k, len(diff)))
        #    else:
        #        d = scores[ix] - mean_amomaly / max_min
        #        step = self.random_state.uniform(low=0, high=1.0-d, size=(k, len(diff)))
        #    for ii in range(k):
        #        samples.append(X[ix] + step[ii,:].ravel()*diff)

        #return np.vstack(samples)

        return samples

    def _bus_generator(self, X, n_to_sample, scores):
        """
        Make samples with bus topology

        Args:
            X (np.array): the feature vectory
            n_to_sample (int): the number of samples to generate
            scores (np.array): classification scores

        Returns:
            np.array: the generated feature vectors
        """

        max_min = max(scores) - min(scores)

        # this array stores all generated samples
        samples = np.zeros(shape=(n_to_sample, X.shape[1]))

        # the base points
        sample_indices = self.random_state.choice(X.shape[0]-1, n_to_sample, replace=True)
        base_points = X[1:][sample_indices]
        diffs = X[:-1] - X[1:]
        diffs_all = diffs[sample_indices]
        diffs_scores = scores[:-1] - scores[1:]
        diffs_scores_all = diffs_scores[sample_indices]

        greater_equal_mask = diffs_scores_all >= 0
        less_mask = ~greater_equal_mask

        d_ge = diffs_scores_all[greater_equal_mask]

        if max_min > 0:
            d_ge = diffs_scores_all[greater_equal_mask] / max_min

        steps_ge = self.random_state.uniform(size=(np.sum(greater_equal_mask), X.shape[1]))
        steps_ge = (steps_ge.T * (1.0 - d_ge) + d_ge).T

        samples[greater_equal_mask] = base_points[greater_equal_mask] + steps_ge * diffs_all[greater_equal_mask]

        d_less = -diffs_scores_all[less_mask]

        if max_min > 0:
            d_less = -diffs_scores_all[less_mask] / max_min

        steps_less = self.random_state.uniform(size=(np.sum(less_mask), X.shape[1]))
        steps_less = (steps_less.T * (1.0-d_less)).T

        samples[less_mask] = base_points[less_mask] + steps_less * diffs_all[less_mask]

        return samples

        # Implementation of the bus topology
        #samples = []
        #
        #k = max([1, int(np.rint(n_to_sample/len(X)))])
        #max_min = max(scores) - min(scores)
        #for ix in range(1, len(X)):
        #    diff = X[ix-1] - X[ix]
        #    if scores[ix-1]>=scores[ix]:
        #        d = (scores[ix-1]-scores[ix]) / max_min
        #        step = 1.0 * self.random_state.uniform(low=0+d, high=1.0, size=(k, len(diff)))
        #    else:
        #        d = (scores[ix]-scores[ix-1]) / max_min
        #        step = 1.0 * self.random_state.uniform(low=0, high=1.0-d, size=(k, len(diff)))
        #    for ii in range(k):
        #        samples.append(X[ix] + step[ii]*diff)

        #return np.vstack(samples)

    def _mesh_generator(self, X, n_to_sample, scores):
        """
        Make samples with mesh topology

        Args:
            X (np.array): the feature vectory
            n_to_sample (int): the number of samples to generate
            scores (np.array): classification scores

        Returns:
            np.array: the generated feature vectors
        """

        # Implementation of the mesh topology
        max_min = max(scores) - min(scores)

        # this array stores all generated samples
        samples = np.zeros(shape=(n_to_sample, X.shape[1]))

        # the base points
        sample_indices = self.random_state.choice(X.shape[0], n_to_sample, replace=True)
        other_indices = self.random_state.choice(X.shape[0], n_to_sample, replace=True)
        base_points = X[sample_indices]
        other_points = X[other_indices]

        diffs = base_points - other_points
        diffs_scores = scores[sample_indices] - scores[other_indices]

        greater_equal_mask = diffs_scores >= 0
        less_mask = ~greater_equal_mask

        d_ge = diffs_scores[greater_equal_mask]

        if max_min > 0:
            d_ge = diffs_scores[greater_equal_mask] / max_min

        steps_ge = self.random_state.uniform(size=(np.sum(greater_equal_mask), X.shape[1]))
        steps_ge = (steps_ge.T * (1.0 - d_ge) + d_ge).T

        samples[greater_equal_mask] = other_points[greater_equal_mask] + steps_ge * diffs[greater_equal_mask]

        d_less = -diffs_scores[less_mask]

        if max_min > 0:
            d_less = -diffs_scores[less_mask] / max_min

        steps_less = self.random_state.uniform(size=(np.sum(less_mask), X.shape[1]))
        steps_less = (steps_less.T * (1.0-d_less)).T

        samples[less_mask] = other_points[less_mask] + steps_less * diffs[less_mask]

        return samples

        # Implementation of the mesh topology
        #samples = []

        #Nx = X.shape[0]
        #if Nx < 3:
        #    n_combs = Nx
        #else:
        #    n_combs = Nx + 0.5*(Nx*(Nx-3))
        #k = max([1, int(np.rint(n_to_sample/n_combs))])
        #if k > 1:
        #    max_min = max(scores) - min(scores)
        #    for i in range(Nx):
        #        for j in range(i+1, Nx):
        #            diff = X[i] - X[j]
        #            if scores[i]>=scores[j]:
        #                d = (scores[i]-scores[j]) / max_min
        #                step = 1.0 * self.random_state.uniform(low=0+d, high=1.0, size=(k, len(diff)))
        #            else:
        #                d = (scores[j]-scores[i]) / max_min
        #                step = 1.0 * self.random_state.uniform(low=0, high=1.0-d, size=(k, len(diff)))
        #            for ii in range(k):
        #                samples.append(X[j] + step[ii,:].ravel()*diff)
        #else:
        #    while len(samples) < n_to_sample:
        #        random_i = self.random_state.randint(len(X))
        #        random_j = self.random_state.randint(len(X))
        #        diff = X[random_i] - X[random_j]
        #        max_min = max(scores) - min(scores)
        #        if scores[random_i]>=scores[random_j]:
        #            d = (scores[random_i]-scores[random_j]) / max_min
        #            step = 1.0 * self.random_state.uniform(low=0+d, high=1.0, size=(k, len(diff)))
        #        else:
        #            d = (scores[random_j]-scores[random_i]) / max_min
        #            step = 1.0 * self.random_state.uniform(low=0, high=1.0-d, size=(k, len(diff)))
        #        for ii in range(k):
        #            samples.append(X[random_j] + step[ii,:].ravel()*diff)

        #return np.vstack(samples)

    def _make_samples(self, X, n_to_sample, scores, classifier):
        """
        Make samples

        Args:
            X (np.array): the feature vectory
            n_to_sample (int): the number of samples to generate
            scores (np.array): classification scores
            classifier (obj): the classifier

        Returns:
            np.array: the generated feature vectors
        """
        samples = None
        if self.topology=='star':
            samples = self._star_generator(X, n_to_sample, scores, classifier)
        elif self.topology=='bus':
            samples = self._bus_generator(X, n_to_sample, scores)
        elif self.topology=='mesh':
            samples = self._mesh_generator(X, n_to_sample, scores)

        return samples

    def _in_danger_iforest(self, X, y, class_label):
        """
        Check if samples are indanger

        Args:
            X (np.array): feature vectors
            y (np.array): the target labels
            class_label (int): the class label to check

        Returns:
            np.array, clf: the class scores and the classifier
        """

        clf = IsolationForest(n_estimators=100, random_state=self._random_state_init)
        clf.fit(X)

        X_scores = clf.decision_function(X)
        class_scores = X_scores[y==class_label]

        return class_scores, clf

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        _logger.info(self.__class__.__name__ + ": " +
                     "Running sampling via %s" % self.descriptor())

        X_min = X[y == self.min_label]

        n_to_sample = self.det_n_to_sample(self.proportion,
                                           self.class_stats[self.maj_label],
                                           self.class_stats[self.min_label])
        if n_to_sample == 0:
            _logger.warning(self.__class__.__name__ + ": " + "Sampling is not needed")
            return X, y

        mda_scores, clf_anomaly = self._in_danger_iforest(X, y, self.min_label)

        X_new = self._make_samples(X_min,
                                    n_to_sample,
                                    mda_scores,
                                    clf_anomaly)

        return np.vstack([X, X_new]), np.hstack([y, np.repeat(self.min_label, X_new.shape[0])])

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'topology': self.topology,
                'random_state': self._random_state_init}


"""
Tests the configurations
"""

from smote_variants.config import (suppress_external_warnings,
                                   suppress_internal_warnings,
                                   distance_matrix_max_memory_chunk,
                                   DEFAULT_SUPPRESS_EXTERNAL_WARNINGS,
                                   DEFAULT_SUPPRESS_INTERNAL_WARNINGS,
                                   DISTANCE_MATRIX_MAX_MEMORY_CHUNK)

def test_external_warnings():
    """
    Tests the external warnings settings
    """

    suppress_external_warnings(True)

    assert suppress_external_warnings()

    suppress_external_warnings(False)

    assert not suppress_external_warnings()

    suppress_external_warnings(DEFAULT_SUPPRESS_EXTERNAL_WARNINGS)

def test_internal_warnings():
    """
    Tests the internal warnings settings
    """

    suppress_internal_warnings(True)

    assert suppress_internal_warnings()

    suppress_internal_warnings(False)

    assert not suppress_internal_warnings()

    suppress_internal_warnings(DEFAULT_SUPPRESS_INTERNAL_WARNINGS)

def test_distance_matrix_memory():
    """
    Testing the setting of the distance matrix memory chunk size
    """

    assert distance_matrix_max_memory_chunk() == DISTANCE_MATRIX_MAX_MEMORY_CHUNK

    distance_matrix_max_memory_chunk(10_000_000)

    assert distance_matrix_max_memory_chunk() == 10_000_000

    distance_matrix_max_memory_chunk(50_000_000)

    assert distance_matrix_max_memory_chunk() == 50_000_000



"""
Testing the SMOTEWB.
"""

import numpy as np

from smote_variants import SMOTEWB


def all_noise_by_dtrees():
    """
    A data set where class membership cannot be judged based on the features.
    Different decision trees can easily provide different results.
    """
    X = np.array([[i, 80 - i] for i in range(0, 80)])
    y = np.zeros(X.shape[0], int)
    y[15] = 1
    y[10] = 1
    y[5] = 1
    return X, y


def low_kmax_good_noise():
    """ Dataset with sample that is considered noise by decision trees,
    but classified as a minor by 1NN, and result low k_max value."""
    X = np.array([
        [0, 0], [0, 1], [0, 2], [0, 3], [0, 4],
        [1, 0], [1, 1], [1, 2], [1, 3], [1, 4],
        [2, 0], [2, 1], [2, 2], [2, 3], [2, 4],
        [3, 0], [3, 1], [3, 2], [3, 3], [3, 4],
        [4, 0], [4, 1], [4, 2], [4, 3], [4, 4]])

    y = np.array([1, 1, 0, 0, 1,
                  0, 0, 0, 0, 0,
                  0, 0, 0, 0, 0,
                  0, 0, 1, 1, 1,
                  1, 0, 1, 1, 1])

    return X, y


def test_specific():
    """
    Oversampler specific testing
    """

    obj = SMOTEWB(n_iters=30, max_depth=3, proportion=0.25)

    X, y = all_noise_by_dtrees()
    X_samp, _ = obj.sample(X, y)

    assert len(X_samp) >= len(X)

    obj = SMOTEWB(n_iters=100, max_depth=3, proportion=0.75)

    X, y = low_kmax_good_noise()
    X_samp, _ = obj.sample(X, y)

    assert len(X_samp) >= len(X)



"""
This module tests the parallelization ecosystem.
"""

import time
import threading
import multiprocessing
from queue import Empty

import pytest

from smote_variants.evaluation import (TimeoutJobBase,
                                        ThreadTimeoutProcessPool,
                                        wait_for_lock,
                                        queue_get_default)

sleeps = [1, 2, 6, 7]

sleeps_no_timeout = [1, 2]

class SleepJob(TimeoutJobBase):
    """
    Sleep job for testing
    """
    def __init__(self, sleep):
        """
        The constructor of the sleep job
        """
        self.sleep = sleep

    def execute(self):
        """
        Execute the sleep job

        Returns:
            dict: the result of the sleep job
        """
        time.sleep(self.sleep)
        return {'slept': self.sleep}

    def timeout(self):
        """
        Carry out the timeout process for the sleep job

        Returns:
            dict: the return of the timeout post-processing
        """
        return {'slept': None}

def sleep_job(sleep):
    """
    Sleep job function

    Args:
        sleep (float): the time to sleep

    Returns:
        dict: the result of the sleeping
    """
    time.sleep(sleep)
    return {'slept': sleep}

def test_jobs_objects_timeout():
    """
    Testing the job objects with timeout.
    """

    ttpp = ThreadTimeoutProcessPool(n_jobs=2, timeout=4)

    results = ttpp.execute([SleepJob(sleep) for sleep in sleeps])

    assert len(results) == len(sleeps)

    assert all(isinstance(result, dict) for result in results)

def test_jobs_functions_timeout():
    """
    Testing the functions with timeout.
    """

    ttpp = ThreadTimeoutProcessPool(n_jobs=2, timeout=4)

    results = ttpp.apply(sleep_job, sleeps)

    assert len(results) == len(sleeps)

    dict_count = sum(isinstance(result, dict) for result in results)

    assert 0 < dict_count < len(sleeps)

def test_jobs_objects_no_timeout():
    """
    Testing the job objects without timeout.
    """

    ttpp = ThreadTimeoutProcessPool(n_jobs=2, timeout=-1)

    results = ttpp.execute([SleepJob(sleep) for sleep in sleeps_no_timeout])

    assert len(results) == len(sleeps_no_timeout)

    assert all(isinstance(result, dict) for result in results)

def test_jobs_functions_no_timeout():
    """
    Testing the job functions without timeout.
    """

    ttpp = ThreadTimeoutProcessPool(n_jobs=2, timeout=-1)

    results = ttpp.apply(sleep_job, sleeps_no_timeout)

    assert len(results) == len(sleeps_no_timeout)

    dict_count = sum(isinstance(result, dict) for result in results)

    assert dict_count == len(sleeps_no_timeout)

def test_exceptions():
    """
    Testing the exception in the base class
    """

    toj = TimeoutJobBase()

    with pytest.raises(RuntimeError) as _:
        toj.execute()

    with pytest.raises(RuntimeError) as _:
        toj.timeout()

def test_lock():
    """
    Testing the wait for lock functionality.
    """
    lock = multiprocessing.Lock()
    lock.acquire()

    thread = threading.Thread(target=wait_for_lock, args=(lock,))
    thread.start()

    time.sleep(3)

    lock.release()

    thread.join(1)

    time.sleep(1)

    assert not thread.is_alive()

class MockQueue: # pylint: disable=too-few-public-methods
    """
    Class mocking the Queue object
    """
    def get(self, block):
        """
        Mocking the get function of the Queue

        Args:
            block (bool): whether to block

        Returns:
            obj: the content of the queue
        """
        _ = block
        raise Empty

class MockJob: # pylint: disable=too-few-public-methods
    """
    Class mocking a TimeoutJob object
    """
    def timeout(self):
        """
        The timeout function

        Returns:
            int: 1
        """
        return 1

def test_queue_get_default():
    """
    Testing the queue_get_default functionality.
    """
    assert queue_get_default(MockQueue(), MockJob()) == 1



"""
A module for global package configurations.
"""

__all__=['suppress_external_warnings',
         'suppress_internal_warnings',
         'distance_matrix_max_memory_chunk',
         'DEFAULT_SUPPRESS_EXTERNAL_WARNINGS',
         'DEFAULT_SUPPRESS_INTERNAL_WARNINGS',
         'DISTANCE_MATRIX_MAX_MEMORY_CHUNK']

DEFAULT_SUPPRESS_EXTERNAL_WARNINGS = True
DEFAULT_SUPPRESS_INTERNAL_WARNINGS = True
DISTANCE_MATRIX_MAX_MEMORY_CHUNK=50_000_000

flags = {'SUPPRESS_EXTERNAL_WARNINGS': DEFAULT_SUPPRESS_EXTERNAL_WARNINGS,
         'SUPPRESS_INTERNAL_WARNINGS': DEFAULT_SUPPRESS_INTERNAL_WARNINGS,
         'DISTANCE_MATRIX_MAX_MEMORY_CHUNK': DISTANCE_MATRIX_MAX_MEMORY_CHUNK}

def suppress_external_warnings(flag=None):
    """
    Set/get the value of SUPPRESS_EXTERNAL_WARNINGS

    Args:
        flag (bool): the new flag

    Returns:
        bool: the current value of the flag
    """

    if flag is not None:
        flags['SUPPRESS_EXTERNAL_WARNINGS'] = flag
    return flags['SUPPRESS_EXTERNAL_WARNINGS']

def suppress_internal_warnings(flag=None):
    """
    Set/get the value of SUPPRESS_INTERNAL_WARNINGS

    Args:
        flag (bool): the new flag

    Returns:
        bool: the current value of the flag
    """

    if flag is not None:
        flags['SUPPRESS_INTERNAL_WARNINGS'] = flag
    return flags['SUPPRESS_INTERNAL_WARNINGS']

def distance_matrix_max_memory_chunk(value=None):
    """
    Set/get the maximum memory chunk for computing distance matrices

    Args:
        value (int): the maximum size of memory to be used

    Returns:
        int: the current value
    """

    if value is not None:
        flags['DISTANCE_MATRIX_MAX_MEMORY_CHUNK']= value
    return flags['DISTANCE_MATRIX_MAX_MEMORY_CHUNK']



"""
This module implements the SMOTEWB method.
"""
import logging
import math

import numpy as np
from sklearn.preprocessing import MinMaxScaler
from sklearn.tree import DecisionTreeClassifier

from smote_variants.base import OverSampling, NearestNeighborsWithMetricTensor, coalesce
from smote_variants.oversampling import SMOTE

_logger = logging.getLogger('smote_variants')

__all__ = ['SMOTEWB']


class SMOTEWB(OverSampling):
    """
    References:
        * BibTex::
           @article{SAGLAM2022117023,
                title = {A novel SMOTE-based resampling technique trough noise
                detection and the boosting procedure},
                journal = {Expert Systems with Applications},
                volume = {200},
                pages = {117023},
                year = {2022},
                issn = {0957-4174},
                doi = {https://doi.org/10.1016/j.eswa.2022.117023}}
    """

    categories = [OverSampling.cat_noise_removal,        # applies noise removal
                  OverSampling.cat_uses_classifier,      # uses some advanced classifier
                  OverSampling.cat_sample_ordinary,      # sampling is done in the SMOTE scheme
                  OverSampling.cat_extensive,            # adds minority samples only
                  OverSampling.cat_metric_learning]      # metric learning is applicable (uses nearest neighbors)

    sample_good = 1
    sample_lonely = 0
    sample_bad = -1

    def __init__(self,
                 proportion=1.0,
                 *,
                 n_iters=100,
                 max_depth=30,
                 nn_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the SMOTE template object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_iters (int): number of iterations (number of weak classifiers, dtrees) of the ensemble
                                noise filtering process (in the paper: M)
            max_depth (int): maximum depth of dtrees (>=3)
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (None/int/np.random.RandomState): the random state
            _kwargs: for technical reasons and to facilitate serialization, additional
                     keyword arguments are accepted
        """

        OverSampling.__init__(self, random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_iters, "n_iters", 1)
        self.check_greater_or_equal(max_depth, "max_depth", 3)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_iters = n_iters
        self.max_depth = max_depth
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.
        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_iters': [30, 50, 100],
                                  'max_depth': [5, 10, 20, 30]}

        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def boosted_weights(self, X, y):
        """
        Implementation of Algorithm 2.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.array): weight of the samples
        """
        n_samples = X.shape[0]
        w = np.repeat(1.0/n_samples, n_samples)

        # based on boosted_weights.R
        clf = DecisionTreeClassifier(random_state=self._random_state_init,
                                     max_depth=self.max_depth,
                                     min_samples_split=3)

        for _ in range(0, self.n_iters):
            clf.fit(X, y, sample_weight=w)
            predicted_labels = clf.predict(X)

            w_error = sum(w[predicted_labels != y])

            # no zero check in the original code (samples are not_noise)
            if w_error == 0:
                return np.zeros(n_samples, float)

            alpha = 0.5 * np.log((1 - w_error) / w_error)
            w[predicted_labels != y] = w[predicted_labels != y] * np.exp(alpha)

            w = w / sum(w)

        return w

    @staticmethod
    def noise_threshold(w, th):
        noise = np.repeat(False, len(w))
        noise[w > th] = True
        return noise

    def noise_detection(self, X, y):
        """
        Implementation of Algorithm 3.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.array, np.array): Mask of noise for the minority and the majority samples, respectively.
        """

        w = self.boosted_weights(X, y)

        w_min = w[y == self.min_label]
        w_maj = w[y == self.maj_label]

        n_samples = len(y)

        # noise thresholds for majority and minority classes
        th_min = (2 * len(w_maj)) / (n_samples ** 2)
        th_maj = (2 * len(w_min)) / (n_samples ** 2)

        noise_mask_min = SMOTEWB.noise_threshold(w_min, th_min)
        noise_mask_maj = SMOTEWB.noise_threshold(w_maj, th_maj)

        return noise_mask_min, noise_mask_maj

    def first_majority_index(self, nearest_ys, no_majority_result=-1):
        """
        Finds the index of the first majority element in the nearest_ys per row.

        Args:
            nearest_ys (np.ndarray): label of the nearest neighbours
            no_majority_result (int): result if there is no majority item in a row
        Returns:
            np.array: the i-th element of the array is the index of the first majority label
                        in the i-th row of the nearest_ys, or no_majority_result if the row
                        contains no majority labels.
        """
        mask = (nearest_ys == self.maj_label)
        return np.where(mask.any(axis=1), mask.argmax(axis=1), no_majority_result)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels
        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        # determine the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        # use logging in the below format
        message = "sampling"
        _logger.info(f"{self.__class__.__name__}: {message}")

        # Step 1: MinMaxScaler - Rescaling is ignored at the suggestion of the author of the algorithm.  

        # Step2: separating the classes
        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        # Step 3-4: creating masks of noise samples
        noise_mask_min, noise_mask_maj = self.noise_detection(X, y)

        # Start: Algorithm 4
        n_min = len(X_min)
        n_maj = len(X_maj)

        # paper: floor, R implementation: ceil
        k_max = math.floor(n_maj / n_min)

        # non-noise samples and labels
        X_min_not_noise = X_min[~noise_mask_min]
        X_maj_not_noise = X_maj[~noise_mask_maj]
        X_not_noise = np.concatenate([X_min_not_noise, X_maj_not_noise])
        y_not_noise = np.concatenate([np.repeat(self.min_label, X_min_not_noise.shape[0]),
                                      np.repeat(self.maj_label, X_maj_not_noise.shape[0])])

        # fitting the model
        # paper: X^good_pos makes no sense, R code: x_notnoise
        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = self.metric_tensor_from_nn_params(nn_params, X, y)

        # all data points are noise
        if len(X_not_noise) == 0:
            smote = SMOTE(proportion=self.proportion, nn_params=self.nn_params)
            return smote.sample(X, y)

        n_neighbors = min([len(X_not_noise), k_max + 1])
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)

        nnmt.fit(X_not_noise)
        indices = nnmt.kneighbors(X_min, n_neighbors=n_neighbors, return_distance=False)

        # number of the positive neighbors until the first negative one (per row)
        k_arr = self.first_majority_index(y_not_noise[indices], no_majority_result=k_max+1)

        # decrease one if the sample is not a noise
        k_arr[~noise_mask_min] -= 1
        # in case there is a majority sample at 0 distance
        k_arr[k_arr == -1] = 0
        # all neighbours are positive
        k_max = min(indices.shape[1], k_max)
        k_arr[k_arr > k_max] = k_max

        # setting the labels of the samples of X_pos
        fl_arr = np.empty(n_min, int)
        fl_arr[k_arr > 0] = SMOTEWB.sample_good
        fl_arr[(k_arr == 0) & noise_mask_min] = SMOTEWB.sample_bad
        fl_arr[(k_arr == 0) & (~noise_mask_min)] = SMOTEWB.sample_lonely
        # End: Alg 4

        # Step 5: n_to_sample - done
        # Step 6: computing n_to_sample per samples
        n_min = len(X_min)

        # number of good and lonely samples
        n_good_min = np.sum(fl_arr == SMOTEWB.sample_good)
        n_lonely_min = np.sum(fl_arr == SMOTEWB.sample_lonely)

        # prevent errors (missing from the original) no good or lonely samples => give chance to the others
        if n_good_min + n_lonely_min == 0:
            smote = SMOTE(proportion=self.proportion, nn_params=self.nn_params)
            return smote.sample(X, y)

        n_to_sample_per_sample = math.ceil(n_to_sample / (n_good_min + n_lonely_min))
        C = np.zeros(n_min, int)
        C[fl_arr == SMOTEWB.sample_good] = n_to_sample_per_sample
        C[fl_arr == SMOTEWB.sample_lonely] = n_to_sample_per_sample

        # Step 8: correcting the number of samples to be generated to achieve the desired balance
        # paper: ceil => n_to_sample-p.sum(C) zero or negative
        # we probably have too many samples because of the ceil
        diff = np.sum(C) - n_to_sample
        if diff > 0:
            good_and_lonely_ind = np.where((fl_arr == SMOTEWB.sample_good) | (fl_arr == SMOTEWB.sample_lonely))[0]
            selected_ind = self.random_state.choice(good_and_lonely_ind, diff, replace=len(good_and_lonely_ind) < diff)
            C[selected_ind] -= 1

        # Step 9: sample generation
        synt_sample_list = []
        for i in range(0, n_min):
            if fl_arr[i] == SMOTEWB.sample_lonely:
                for j in range(C[i]):
                    synt_sample_list.append(X_min[i].copy())
            elif fl_arr[i] == SMOTEWB.sample_good and C[i] > 0:
                if noise_mask_min[i]:
                    nn = indices[i, 0: k_arr[i]]
                else:
                    # removing self index
                    nn = indices[i, 1: k_arr[i] + 1]

                if len(nn) > 0:
                    k_ids = self.random_state.choice(nn, C[i])
                    for j in k_ids:
                        synt_sample_list.append(self.sample_between_points(X_min[i], X_not_noise[j]))

        # Step 11-12: merging the original samples and the synthetic ones
        X_synt_samples = np.array(synt_sample_list)
        X_samples = np.vstack([X, X_synt_samples])

        return (X_samples,
                np.hstack([y, np.hstack([self.min_label]*len(synt_sample_list))]))

    def preprocessing_transform(self, X):
        """
        Scale the attributes of X into [0, 1] range.

        Args:
            X (np.array): features
        Returns:
            np.array: transformed features
        """
        return X

    def get_params(self, deep=False):
        """
        Returns the parameters of the object

        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'max_depth': self.max_depth,
                'nn_params': self.nn_params,
                'n_iters': self.n_iters,
                **OverSampling.get_params(self, deep)}
